package editor;

import dicewars.model.Hex;
import dicewars.file.FileService;
import dicewars.util.Point;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.event.Event;
import javafx.event.EventHandler;
import javafx.scene.Scene;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.control.*;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.FlowPane;
import javafx.scene.paint.Color;
import javafx.scene.text.Font;
import javafx.stage.FileChooser;
import javafx.stage.Stage;

import java.io.File;
import java.io.IOException;
import java.util.*;

import static java.lang.System.out;

/*
 *  Crude utility to create Hex grids
 *
 *  Use to create yout own grad maps
 *  No warranties what so ever ....
 *
 *  *** Nothing to do here ***
 */
public class HexGridEditor extends Application {

    // A list of hexes belonging to some color
    private final Map<Color, List<Hex>> regions = new HashMap<>();
    private int gridSize;

    // --- Mouse Events ----------------
    private void selectHex(MouseEvent me) {

        Point p = new Point(me.getX()-margin, me.getY()-margin);
        Hex h = new Hex(p, hexRadius);
        //out.println(h);

        // Move selected Hex from actual color list to other list
        for (Map.Entry<Color, List<Hex>> e : regions.entrySet()) {
            int i = e.getValue().indexOf(h);
            // If mouse position not outside grid
            if (i >= 0) {
                // Remove hex from actual list
                Hex hex = e.getValue().remove(i);
                // Select color list to store removed hex
                List<Hex> dest = regions.get(colorPicker.getValue());
                // If list exists add, else create list and add
                if (dest != null) {
                    dest.add(hex);
                } else {
                    List<Hex> region = new ArrayList<>();
                    region.add(hex);
                    regions.put(colorPicker.getValue(), region);
                }
                break;
            }
        }
        renderMap(hexRadius);
    }

    // Possible, but not used.
    /*private void render(MouseEvent me) {
        renderMap(hexRadius);
    }*/

    //------ Menu actions  ----------------------

    private void newMap(ActionEvent e) {
        List<Integer> choices = new ArrayList<>();
        choices.add(6);
        choices.add(8);
        choices.add(10);
        choices.add(12);
        choices.add(14);
        choices.add(16);
        ChoiceDialog<Integer> dialog = new ChoiceDialog<>(8, choices);
        dialog.setTitle("New Map");
        dialog.setHeaderText("Select Map Size");
        dialog.setContentText("Choose size:");
        Optional<Integer> result = dialog.showAndWait();
        if (result.isPresent()) {
            gridSize = result.get();
            hexRadius = 0.5 * width / gridSize;
            margin = hexRadius;
            regions.clear();
            List<Hex> hexes = generateHexGrid(result.get());
            regions.put(Color.WHITE, hexes);
            colorPicker.setValue(Color.WHITE);
            renderMap(hexRadius);
        }
    }

    private void save(Event e) {
        FileChooser fileChooser = new FileChooser();
        fileChooser.setTitle("Save Hex Map");
        fileChooser.setInitialFileName("hexmap.txt");
        File file = fileChooser.showSaveDialog(stage);
        if (file != null) {
            try {
                FileService.writeHexGrid(file.toPath(), regions);
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
    }

    private void exit(Event e) {
        System.exit(0);
    }

    // ------ Utility ------------

    private List<Hex> generateHexGrid(int side) {
        List<Hex> hexes = new ArrayList<>();
        for (int yh = 0; yh < side; yh++) {
            for (int xh = - yh/2; xh < side - yh/2; xh++) {
                hexes.add(new Hex(xh, yh));
            }
        }
        return hexes;
    }

    // ------ GUI and JavaFX Stuff ------------------------------

    private final int height = 600;
    private final int width = 600;

    private double hexRadius;  // Will be set in newMap()
    private final ColorPicker colorPicker = new ColorPicker();
    private GraphicsContext gc;
    private Stage stage;

    @Override
    public void start(Stage primaryStage) throws Exception {

        stage = primaryStage;
        BorderPane root = new BorderPane();

        // Menu
        MenuBar menu = getMenu();
        menu.prefWidthProperty().bind(primaryStage.widthProperty());
        root.setTop(menu);

        // Toolbar
        FlowPane toolBar = new FlowPane();
        toolBar.setHgap(10);
        toolBar.setStyle("-fx-border-color: black; -fx-spacing: 10; -fx-padding: 10");

        // Color picker
        Label l = new Label("Same color will belong to same region (even if separated)");
        toolBar.getChildren().addAll(l, colorPicker);
        root.setBottom(toolBar);

        // Canvas
        Canvas canvas = new Canvas(width, height);
        canvas.setOnMouseDragged(this::selectHex);
        //canvas.setOnMouseReleased(this::render);
        gc = canvas.getGraphicsContext2D();
        root.setCenter(canvas);

        // Scene
        Scene scene = new Scene(root);
        primaryStage.setScene(scene);
        primaryStage.setTitle("Map Editor");
        primaryStage.show();

    }

    // --- Menu ---------
    private MenuBar getMenu() {

        Menu menuFile = new Menu("File");
        MenuItem[] menuFileItems = {
                new MenuItemExt("New Map", this::newMap),
                new MenuItemExt("Save", this::save),
                new MenuItemExt("Quit", this::exit)
        };
        menuFile.getItems().addAll(menuFileItems);

        // -- The bar for all menus
        MenuBar menuBar = new MenuBar();
        menuBar.getMenus().addAll(menuFile);
        return menuBar;
    }

    //  Helper class with more useful constructor
    private static class MenuItemExt extends MenuItem {
        public MenuItemExt(String label, EventHandler<ActionEvent> e) {
            super(label);
            setOnAction(e);
        }
    }

    // ----- Rendering -----------------------------
    private double margin;
    private void renderMap(double hexRadius) {
        gc.clearRect(0, 0, width, height);
        for (Map.Entry<Color, List<Hex>> e : regions.entrySet()) {
            for (Hex h : e.getValue()) {
                renderHex(h, e.getKey(), hexRadius);
            }
        }

    }

    private void renderHex(Hex hex, Color color, double radius) {
        double[] xs = new double[6];
        double[] ys = new double[6];
        getXsAndYs(hex.getCorners(radius), xs, ys);
        gc.setFill(color);
        gc.fillPolygon(xs, ys, 6);
        gc.setStroke(Color.BLACK);
        gc.strokePolygon(xs, ys, 6);
        // Visualize coordinates
        Point p = hex.toPoint(hexRadius);
        Font font = Font.font(10);
        gc.setFont(font);
        gc.strokeText(hex.getXh() + ":" + hex.getYh(), p.x + margin - 10, p.y + margin);
    }

    // Convert List<Point> to parallel arrays of coordinates
    private void getXsAndYs(List<Point> pts, double[] xs, double[] ys) {
        int i = 0;
        for (Point p : pts) {
            xs[i] = p.x + margin;
            ys[i] = p.y + margin;
            i++;
        }
    }

    public static void main(String[] args) {
        launch();
    }

}
